//
// Copyright (c) 2015-2019 CNRS INRIA
// Copyright (c) 2015 Wandercraft, 86 rue de Paris 91400 Orsay, France.
//

#ifndef __pinocchio_multibody_visitor_hpp__
#define __pinocchio_multibody_visitor_hpp__

#include "pinocchio/multibody/visitor/joint-unary-visitor.hpp"
#include "pinocchio/multibody/visitor/joint-binary-visitor.hpp"

#endif // ifndef __pinocchio_multibody_visitor_hpp__
